

#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging

from datetime import datetime

from src.base.config import AuraConfig
from src.base.utils import SimpleUtil as SU
from src.core.engine import Engine

from src.scheduling.calendar import AuraCalendarService
from src.scheduling.models import Timeslot




class Programme():
    """
    The current programme of the calendar. The programme is consisting of a set of timeslots.
    """
    config = None
    logger = None
    programme = None
    last_successful_fetch = None


    def __init__(self):
        """
        Constructor
        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")



    def refresh(self):
        """
        Fetch the latest programme from `AuraCalendarService` which stores it to the database.
        After that, the programme is in turn loaded from the database and stored in `self.programme`.
        """

        # Fetch programme from API endpoints
        self.logger.debug("Trying to fetch new programe from API endpoints...")
        acs = AuraCalendarService(self.config)
        queue = acs.get_queue()
        acs.start() # start fetching thread
        response = queue.get() # wait for the end
        self.logger.debug("... Programme fetch via API done!")

        # Reset last successful fetch state
        lsf = self.last_successful_fetch
        self.last_successful_fetch = None

        if response is None:
            msg = SU.red("Trying to load programme from Engine Database, because AuraCalendarService returned an empty response.")
            self.logger.warning(msg)
        elif type(response) is list:
            self.programme = response
            if self.programme is not None and len(self.programme) > 0:
                self.last_successful_fetch = datetime.now()
                self.logger.info(SU.green("Finished fetching current programme from API"))
            if len(self.programme) == 0:
                self.logger.critical("Programme fetched from Steering/Tank has no entries!")
        elif response.startswith("fetching_aborted"):
            msg = SU.red("Trying to load programme from database only, because fetching was being aborted from AuraCalendarService! Reason: ")
            self.logger.warning(msg + response[16:])
        else:
            msg = SU.red("Trying to load programme from database only, because of an unknown response from AuraCalendarService: " + response)
            self.logger.warning(msg)

        # Always load latest programme from the database
        self.last_successful_fetch = lsf
        self.load_programme_from_db()
        self.logger.info(SU.green("Finished loading current programme from database (%s timeslots)" % str(len(self.programme))))
        for timeslot in self.programme:
            self.logger.debug("\tTimeslot %s with Playlist %s" % (str(timeslot), str(timeslot.playlist)))




    def load_programme_from_db(self):
        """
        Loads the programme from Engine's database and enables
        them via `self.enable_entries(..)`. After that, the
        current message queue is printed to the console.
        """
        self.programme = Timeslot.select_programme()

        if not self.programme:
            self.logger.critical(SU.red("Could not load programme from database. We are in big trouble my friend!"))
            return



    def get_current_entry(self):
        """
        Retrieves the current `PlaylistEntry` which should be played as per programme. 

        Returns:
            (PlaylistEntry): The track which is (or should) currently being played
        """
        now_unix = Engine.engine_time()

        # Load programme if necessary
        if not self.programme:
            self.load_programme_from_db()

        # Check for current timeslot
        current_timeslot = self.get_current_timeslot()
        if not current_timeslot:
            self.logger.warning(SU.red("There's no active timeslot"))
            return None

        # Check for scheduled playlist
        current_playlist = current_timeslot.playlist
        if not current_playlist:
            msg = "There's no playlist assigned to the current timeslot. Most likely a fallback will make things okay again."
            self.logger.warning(SU.red(msg))
            return None

        # Iterate over playlist entries and store the current one
        current_entry = None
        for entry in current_playlist.entries:
            if entry.start_unix <= now_unix and now_unix <= entry.end_unix:
                current_entry = entry
                break
      
        if not current_entry:
            # Nothing playing ... fallback will kick-in
            msg = "There's no entry scheduled for playlist '%s' at %s" % (str(current_playlist), SU.fmt_time(now_unix))
            self.logger.warning(SU.red(msg))
            return None

        return current_entry
 


    def get_current_timeslot(self):
        """
        Retrieves the timeslot currently to be played. 
        
        Returns:
            (Timeslot): The current timeslot
        """
        current_timeslot = None
        now_unix = Engine.engine_time()

        # Iterate over all timeslots and find the one to be played right now
        if self.programme:
            for timeslot in self.programme:
                if timeslot.start_unix <= now_unix and now_unix < timeslot.end_unix:
                    current_timeslot = timeslot
                    break
        
        return current_timeslot



    def get_next_timeslots(self, max_count=0):
        """
        Retrieves the timeslots to be played after the current one.

        Args:
            max_count (Integer): Maximum of timeslots to return, if `0` all exitsing ones are returned

        Returns:
            ([Timeslot]): The next timeslots
        """
        now_unix = Engine.engine_time()
        next_timeslots = []

        for timeslot in self.programme:
            if timeslot.start_unix > now_unix:
                if (len(next_timeslots) < max_count) or max_count == 0:
                    next_timeslots.append(timeslot)
                else:
                    break
                
        return self.filter_scheduling_window(next_timeslots)



    def filter_scheduling_window(self, timeslots):
        """
        Ignore timeslots which are beyond the scheduling window. The end of the scheduling window
        is defined by the config option `scheduling_window_end`. This value defines the seconds
        minus the actual start time of the timeslot.
        """
        now_unix = Engine.engine_time()
        len_before = len(timeslots)
        window_start = self.config.get("scheduling_window_start")
        window_end = self.config.get("scheduling_window_end")
        timeslots = list(filter(lambda s: (s.start_unix - window_end) > now_unix and (s.start_unix - window_start) < now_unix, timeslots))
        len_after = len(timeslots)
        self.logger.info("For now, skipped %s future timeslot(s) which are out of the scheduling window (-%ss <-> -%ss)" % ((len_before - len_after), window_start, window_end))

        return timeslots



    def is_timeslot_in_window(self, timeslot):
        """
        Checks if the timeslot is within the scheduling window.
        """
        now_unix = Engine.engine_time()
        window_start = self.config.get("scheduling_window_start")
        window_end = self.config.get("scheduling_window_end")

        if timeslot.start_unix - window_start < now_unix and \
            timeslot.start_unix - window_end > now_unix:

            return True
        return False