
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import threading
import time
import sqlalchemy

from operator                   import attrgetter
from datetime                   import datetime, timedelta

from src.base.config            import AuraConfig
from src.base.utils             import SimpleUtil as SU
from src.scheduling.models      import AuraDatabaseModel, Timeslot, Playlist
from src.base.exceptions        import NoActiveTimeslotException, LoadSourceException
from src.core.control           import EngineExecutor
from src.core.engine            import Engine
from src.core.channels          import ChannelType, TransitionType, EntryPlayState
from src.core.resources         import ResourceClass, ResourceUtil

from src.scheduling.utils       import TimeslotRenderer
from src.scheduling.programme   import Programme




class TimeslotCommand(EngineExecutor):
    """
    Command for triggering start and end of timeslot events.
    """
    engine = None

    def __init__(self, engine, timeslot):
        """
        Constructor

        Args:
            timeslot (Timeslot):    The timeslot which is starting at this time
        """        
        self.engine = engine

        def do_start_timeslot(timeslot):
            self.logger.info(SU.cyan(f"=== on_timeslot_start('{timeslot}') ==="))
            self.engine.event_dispatcher.on_timeslot_start(timeslot)

        def do_end_timeslot(timeslot):
            self.logger.info(SU.cyan(f"=== on_timeslot_end('{timeslot}') ==="))
            self.engine.event_dispatcher.on_timeslot_end(timeslot)

        child_timer = EngineExecutor("TIMESLOT", None, timeslot.end_unix, do_end_timeslot, timeslot)                            
        super().__init__("TIMESLOT", child_timer, timeslot.start_unix, do_start_timeslot, timeslot)




class AuraScheduler(threading.Thread):
    """
    Aura Scheduler Class

    - Retrieves data from Steering and Tank
    - Executes engine actions in an automated fashion

    Attributes:
        config (AuraConfig):                    Holds the Engine Configuration
        logger:                                 The logger
        exit_event(threading.Event):            Used to exit the thread if requested
        engine:                            Virtual mixer
        last_successful_fetch (datetime):       Stores the last time a fetch from Steering/Tank was successful

        programme:                              The current radio programme to be played as defined in the local engine database
        active_entry(Show, Track):              This is a Tuple consisting of the currently played `Show` and `Track`
        message_timer(List<threading.Timer>):   The timer queue of sound-system commands for playlists/entries to be played
    """
    config = None
    logger = None
    engine = None
    exit_event = None    
    timeslot_renderer = None
    programme = None
    message_timer = []
    fallback = None
    is_initialized = None
    is_initialized = None



    def __init__(self, engine, fallback_manager):
        """
        Constructor

        Args:
            config (AuraConfig):        Reads the engine configuration
            engine (Engine):            The engine to play the schedule on
            func_on_init (Function):    The function to be called when the scheduler is initialized
        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")
        self.programme = Programme()
        self.timeslot_renderer = TimeslotRenderer(self)
        AuraScheduler.init_database()
        self.fallback = fallback_manager
        self.engine = engine
        self.engine.scheduler = self
        self.is_soundsytem_init = False
        
        # Scheduler Initialization
        self.is_initialized = False
        self.is_engine_ready = False

        # Init scheduling thread
        threading.Thread.__init__(self)
        self.exit_event = threading.Event()
        self.start()



    def run(self):
        """
        Called when thread is started via `start()`. It does the following:
        
            1. `self.fetch_new_programme()` periodically from the API depending on the `fetching_frequency` defined in the engine configuration. 
            2. Loads the latest programme from the database and sets the instance state `self.programme` with current timeslots.
            3. Queues all timeslots of the programme, if the soundssystem is ready to accept commands.

        On every cycle the configuration file is reloaded, to allow modifications while running the engine.
        """
        while not self.exit_event.is_set():
            try:
                self.config.load_config()
                seconds_to_wait = int(self.config.get("fetching_frequency"))
                self.logger.info(SU.cyan(f"== start fetching new timeslots (every {seconds_to_wait} seconds) =="))
                
                # Load some stuff from the API in any case
                self.programme.refresh()
                
                # Queue only when the engine is ready to play
                if self.is_initialized == True:                
                    self.queue_programme()

            except Exception as e:
                self.logger.critical(SU.red(f"Unhandled error while fetching & scheduling new programme! ({str(e)})"), e)
                # Keep on working anyway

            self.clean_timer_queue()
            self.print_timer_queue()
                        

            EngineExecutor.log_commands()            
            self.exit_event.wait(seconds_to_wait)



    #
    #   EVENTS
    #


    def on_ready(self):
        """
        Called when the engine has finished booting and is ready to play.
        """
        self.is_initialized = True
        self.on_scheduler_ready()



    def on_scheduler_ready(self):
        """
        Called when the scheduler is ready.
        """
        self.logger.info(self.timeslot_renderer.get_ascii_timeslots())

        try:
            self.play_active_entry()
            self.queue_startup_entries()
        except NoActiveTimeslotException:
            # That's not good, but keep on working...
            pass



    #
    #   METHODS
    #


    def get_programme(self):
        """
        Returns the current programme.
        """
        return self.programme
        

    def play_active_entry(self):
        """
        Plays the entry scheduled for the very current moment and forwards to the scheduled position in time. 
        Usually called when the Engine boots.

        Raises:
            (NoActiveTimeslotException):    If there's no timeslot in the programme, within the scheduling window
        """
        sleep_offset = 10
        active_timeslot = self.programme.get_current_timeslot()

        # Schedule any available fallback playlist
        if active_timeslot:
            # Create command timer to indicate the start of the timeslot
            TimeslotCommand(self.engine, active_timeslot)

            self.fallback.queue_fallback_playlist(active_timeslot)
            # Queue the fade-out of the timeslot
            if not active_timeslot.fadeouttimer:
                self.queue_end_of_timeslot(active_timeslot, True)

        active_entry = self.programme.get_current_entry()
        if not active_entry:
            raise NoActiveTimeslotException

        # In case of a file-system source, we need to fast-foward to the current marker as per timeslot
        if active_entry.get_content_type() in ResourceClass.FILE.types:

            # Calculate the seconds we have to fast-forward
            now_unix = Engine.engine_time()
            seconds_to_seek = now_unix - active_entry.start_unix

            # If the seek exceeds the length of the current track, 
            # there's no need to do anything - the scheduler takes care of the rest
            if (seconds_to_seek + sleep_offset) > active_entry.duration:
                self.logger.info("The FFWD [>>] range exceeds the length of the entry. Drink some tea and wait for the sound of the next entry.")
            else:                
                # Preload and play active entry
                self.engine.player.preload(active_entry)
                self.engine.player.play(active_entry, TransitionType.FADE)

                # Fast-forward to the scheduled position
                if seconds_to_seek > 0:
                    # Without plenty of timeout (10s) the seek doesn't work
                    def async_cue_seek(seconds_to_seek):
                        seconds_to_seek += sleep_offset
                        time.sleep(sleep_offset)
                        self.logger.info("Going to fast-forward %s seconds" % seconds_to_seek)
                        response = self.engine.player.queue_seek(active_entry.channel, seconds_to_seek)
                        self.logger.info("Sound-system seek response: " + response)

                    thread = threading.Thread(target = async_cue_seek, args = (seconds_to_seek,))
                    thread.start()                         

        elif active_entry.get_content_type() in ResourceClass.STREAM.types \
            or active_entry.get_content_type() in ResourceClass.LIVE.types:

            # Preload and play active entry
            self.engine.player.preload(active_entry)
            self.engine.player.play(active_entry, TransitionType.FADE)

        else:
            self.logger.critical("Unknown Entry Type: %s" % active_entry)
        


    def print_timer_queue(self):
        """
        Prints the current timer queue i.e. playlists in the queue to be played.
        """
        message_queue = ""
        messages = sorted(self.message_timer, key=attrgetter('diff'))
        if not messages:
            self.logger.warning(SU.red("There's nothing in the Timer Queue!"))
        else:
            for msg in messages:
                message_queue += str(msg)+"\n"

            self.logger.info("Timer queue: \n" + message_queue) 



    def clean_timer_queue(self):
        """
        Removes inactive timers from the queue.
        """
        len_before = len(self.message_timer)
        self.message_timer[:] = [m for m in self.message_timer if m.is_alive()]
        len_after = len(self.message_timer)
        self.logger.debug("Removed %s finished timer objects from queue" % (len_before - len_after))



    def get_active_playlist(self):
        """
        Retrieves the currently playing playlist.

        Returns:
            (FallbackType, Playlist): The resolved playlist
        """
        timeslot = self.programme.get_current_timeslot()
        if timeslot:
            return self.fallback.resolve_playlist(timeslot)
        return (None, None)



    def queue_programme(self):
        """
        Queues the current programme (playlists as per timeslot) by creating
        timed commands to the sound-system to enable the individual tracks of playlists. 
        """

        # Get a clean set of the timeslots within the scheduling window
        timeslots = self.programme.get_next_timeslots()

        # Queue the timeslots, their playlists and entries
        if timeslots:
            for next_timeslot in timeslots:
                # Create command timer to indicate the start of the timeslot
                TimeslotCommand(self.engine, next_timeslot)

                # Schedule any available fallback playlist
                self.fallback.queue_fallback_playlist(next_timeslot)

                if next_timeslot.playlist:
                    self.queue_playlist_entries(next_timeslot, next_timeslot.playlist.entries, False, True)
                                
                # Queue the fade-out of the timeslot
                if not next_timeslot.fadeouttimer:
                    self.queue_end_of_timeslot(next_timeslot, True)

        self.logger.info(SU.green("Finished queuing programme."))



    def queue_startup_entries(self):
        """
        Queues all entries after the one currently playing upon startup. Don't use
        this method in any other scenario, as it doesn't respect the scheduling window.
        """
        current_timeslot = self.programme.get_current_timeslot()
        
        # Queue the (rest of the) currently playing timeslot upon startup
        if current_timeslot:
            current_playlist = current_timeslot.playlist

            if current_playlist:
                active_entry = self.programme.get_current_entry()

                # Finished entries
                for entry in current_playlist.entries:
                    if active_entry == entry:
                        break

                # Entry currently being played
                if active_entry:

                    # Queue open entries for current playlist
                    rest_of_playlist = active_entry.get_next_entries(True)
                    self.queue_playlist_entries(current_timeslot, rest_of_playlist, False, True)

                    # Store them for later reference
                    current_timeslot.queued_entries = [active_entry]
                    if rest_of_playlist:
                        current_timeslot.queued_entries.append(rest_of_playlist)



    def queue_playlist_entries(self, timeslot, entries, fade_in, fade_out):
        """
        Creates sound-system player commands for all playlist items to be executed at the scheduled time.

        Since each scheduled playlist can consist of multiple entry types such as *file*, *live*,
        and *stream*, the play-out of the timeslot is actually a bit more complex. Before any playlist 
        entries of the timeslot can be turned into sound, they need to be grouped, queued and pre-loaded.

        1. First, all entries are aggregated when they hold filesystem entries.
            Given you have a playlist with 10 entries, the first 4 are consisting of files, the next two
            of a a stream and a live source. The last 4 are files again. These entries are now
            aggregated into 4 groups: one for the files, one for the stream, one for the live entry
            and another one for files. For each group a timer for executing the next step is created.

        2. Now, the playlist entries are going to be "pre-loaded". This means that filesystem
            entries are queued and pre-loaded and entries which are based on audio streams are buffered.
            This is required to allow a seamless play-out, when its time to do so (in the next stage).
            Due to their nature, playlist entries which hold live audio sources are not affected by
            this stage at all.

        Args:
            timeslot (Timeslot):        The timeslot this entries belong to
            entries ([PlaylistEntry]):  The playlist entries to be scheduled for playout
            fade_in (Boolean):          Fade-in at the beginning of the set of entries
            fade_out (Boolean):         Fade-out at the end of the set of entries

        Returns:
            (String):                   Formatted string to display playlist entries in log
        """       
        entry_groups = []
        entry_groups.append([])
        previous_entry = None
        index = 0

        # Mark entries which start after the end of their timeslot or are cut
        # clean_entries = self.preprocess_entries(entries, True)

        # Group/aggregate all filesystem entries, allowing them to be queued at once
        for entry in entries:
            if previous_entry == None or \
                (previous_entry != None and \
                 previous_entry.get_content_type() == entry.get_content_type() and \
                 entry.get_content_type() in ResourceClass.FILE.types):
                
                entry_groups[index].append(entry)
            else:
                index += 1
                entry_groups.append([])
                entry_groups[index].append(entry)
            previous_entry = entry
        self.logger.info("Built %s entry group(s)" % len(entry_groups))
         
        # Timeslot function calls
        do_queue_timeslot_end = False
        if len(entries) > 0 and len(entry_groups) > 0:
            for entries in entry_groups:
                if not isinstance(entries, list):
                    raise ValueError("Invalid Entry Group: %s" % str(entries))

                # Create timers for each entry group
                self.set_entries_timer(entries, fade_in, fade_out)

                # Store them for later reference
                timeslot.queued_entries = entries

        else:
            self.logger.warn(SU.red("Nothing to schedule ..."))



    def set_entries_timer(self, entries, fade_in, fade_out):
        """
        Creates timer for loading and playing one or multiple entries. Existing timers are 
        updated.

        Args:
            entries ([]): List of multiple filesystem entries, or a single entry of other types
        """
        play_timer = self.is_something_planned_at_time(entries[0].start_unix)
        now_unix = Engine.engine_time()
        diff = entries[0].start_unix - now_unix

        # Play function to be called by timer
        def do_play(entries):
            self.logger.info(SU.cyan("=== play('%s') ===" % ResourceUtil.get_entries_string(entries)))
            transition_type = TransitionType.INSTANT
            if fade_in:
                transition_type = TransitionType.FADE

            if entries[-1].status != EntryPlayState.READY:
                self.logger.critical(SU.red("PLAY: For some reason the entry/entries are not yet ready to be played (Entries: %s)" % ResourceUtil.get_entries_string(entries)))
                # Let 'em play anyway ...

            self.engine.player.play(entries[0], transition_type)
            self.logger.info(self.timeslot_renderer.get_ascii_timeslots())


        if play_timer:
            # Check if the Playlist IDs are different
            if self.have_entries_changed(play_timer, entries):
                # If not, stop and remove the old timer, create a new one
                self.stop_timer(play_timer)                
            else:
                # If the playlist entries do not differ => reuse the old timer and do nothing
                self.logger.debug("Playlist Entry %s is already scheduled - no new timer created." % ResourceUtil.get_entries_string(entries))
                return
        
        # If nothing is planned at given time, create a new timer
        (entries[0].switchtimer, entries[0].loadtimer) = self.create_timer(diff, do_play, entries, switcher=True)



    def have_entries_changed(self, timer, new_entries):
        """
        Checks if the new entries and playlists are matching the existing queued ones,
        or if they should be updated.
        
        Args:
            timer (CallFunctionTimer):      The timer holding queued entries
            new_entries ([PlaylistEntry]):  The possibly updated entries

        Returns:
            (Boolean):  `True` if it has changed
        """
        old_entries = timer.entries

        if old_entries[0].playlist and new_entries[0].playlist:
            if old_entries[0].playlist.playlist_id != new_entries[0].playlist.playlist_id:
                return True
        if len(old_entries) != len(new_entries):
            return True

        for old_entry, new_entry in zip(old_entries, new_entries):
            if old_entry.source != new_entry.source:
                return True
        
        return False



    def queue_end_of_timeslot(self, timeslot, fade_out):
        """
        Queues a engine action to stop/fade-out the given timeslot.

        Args:
            timeslot (PlaylistEntry):  The timeslot
            fade_out (Boolean):        If the timeslot should be faded-out
        """
        timeslot_end = timeslot.timeslot_end
        timeslot_end_unix = timeslot.end_unix
        now_unix = Engine.engine_time()
        fade_out_time = 0

        # Stop function to be called when timeslot ends
        def do_stop(timeslot):
            if timeslot.has_queued_entries():
                last_entry = timeslot.queued_entries[-1]
                self.logger.info(SU.cyan("=== stop('%s') ===" % str(last_entry.playlist.timeslot)))
                transition_type = TransitionType.INSTANT
                if fade_out:
                    transition_type = TransitionType.FADE
                self.engine.player.stop(last_entry, transition_type)

        if fade_out == True:
            fade_out_time = int(round(float(self.config.get("fade_out_time")))) #TODO Use float
        
        # Stop any existing fade-out timer
        if timeslot.fadeouttimer:
            timeslot.fadeouttimer.cancel()
            self.message_timer.remove(timeslot.fadeouttimer)

        # Create timer to fade-out
        start_fade_out = timeslot_end_unix - now_unix - fade_out_time
        # last_entry = timeslot.queued_entries[-1]
        timeslot.fadeouttimer = self.create_timer(start_fade_out, do_stop, timeslot, fadeout=True)

        self.logger.info("Fading out timeslot in %s seconds at %s | Timeslot: %s" % (str(start_fade_out), str(timeslot_end), timeslot))



    def is_something_planned_at_time(self, given_time):
        """
        Checks for existing timers at the given time.
        """
        for t in self.message_timer:
            if t.fadein or t.switcher:
                if t.entries[0].start_unix == given_time:
                    return t
        return False



    def create_timer(self, diff, func, param, fadein=False, fadeout=False, switcher=False):
        """
        Creates a new timer for timed execution of mixer commands.

        Args:
            diff (Integer):     The difference in seconds from now, when the call should happen
            func (Function):    The function to call
            param ([]):         A timeslot or list of entries

        Returns:
            (CallFunctionTimer, CallFunctionTimer):     In case of a "switch" command, the switch and preload timer is returned
            (CallFunctionTimer):                        In all other cases only the timer for the command is returned
        """
        if not fadein and not fadeout and not switcher or fadein and fadeout or fadein and switcher or fadeout and switcher:
            raise ValueError("You have to call me with either fadein=true, fadeout=true or switcher=True")
        if not isinstance(param, list) and not isinstance(param, Timeslot):
            raise ValueError("No list of entries nor timeslot passed!")

        t = CallFunctionTimer(diff=diff, func=func, param=param, fadein=fadein, fadeout=fadeout, switcher=switcher)
        self.message_timer.append(t)
        t.start()

        if switcher:
            # Preload function to be called by timer
            def do_preload(entries):
                try:
                    if entries[0].get_content_type() in ResourceClass.FILE.types:
                        self.logger.info(SU.cyan("=== preload_group('%s') ===" % ResourceUtil.get_entries_string(entries)))
                        self.engine.player.preload_group(entries, ChannelType.QUEUE)
                    else:
                        self.logger.info(SU.cyan("=== preload('%s') ===" % ResourceUtil.get_entries_string(entries)))
                        self.engine.player.preload(entries[0])
                except LoadSourceException as e:
                    self.logger.critical(SU.red("Could not preload entries %s" % ResourceUtil.get_entries_string(entries)), e)

                if entries[-1].status != EntryPlayState.READY:
                    self.logger.critical(SU.red("Entries didn't reach 'ready' state during preloading (Entries: %s)" % ResourceUtil.get_entries_string(entries)))

            loader_diff = diff - self.config.get("preload_offset")
            loader = CallFunctionTimer(diff=loader_diff, func=do_preload, param=param, fadein=fadein, fadeout=fadeout, switcher=False, loader=True)
            self.message_timer.append(loader)
            loader.start()
            return (t, loader)
        else:
            return t



    def stop_timer(self, timer):
        """
        Stops the given timer.

        Args:
            timer (Timer):  The timer to stop.
        """
        timer.cancel()
        count = 1

        for entry in timer.entries:
            if entry.loadtimer is not None:
                entry.loadtimer.cancel()
                self.message_timer.remove(entry.loadtimer)
                count += 1

            # if timer.entries[0].fadeintimer is not None:
            #     timer.entries[0].fadeintimer.cancel()
            #     self.message_timer.remove(timer.entries[0].fadeintimer)
            #     count += 1

            # if entry.fadeouttimer is not None:
            #     entry.fadeouttimer.cancel()
            #     self.message_timer.remove(entry.fadeouttimer)
            #     count += 1

        # Remove it from message queue
        self.message_timer.remove(timer)
        self.logger.info("Stopped %s timers for: %s" % (str(count), ResourceUtil.get_entries_string(timer.entries)))



    # FIXME Move to adequate module
    @staticmethod
    def init_database():
        """
        Initializes the database.

        Raises:
            sqlalchemy.exc.ProgrammingError:    In case the DB model is invalid
        """
        if AuraConfig.config().get("recreate_db") is not None:
            AuraDatabaseModel.recreate_db(systemexit=True)

        # Check if tables exists, if not create them
        try:
            Playlist.is_empty()
        except sqlalchemy.exc.ProgrammingError as e:
            errcode = e.orig.args[0]

            if errcode == 1146: # Error for no such table
                model = AuraDatabaseModel()
                model.recreate_db()
            else:
                raise



    def terminate(self):
        """
        Called when thread is stopped or a signal to terminate is received.
        """
        self.exit_event.set()
        self.logger.info("Shutting down scheduler ...")




# ------------------------------------------------------------------------------------------ #
class CallFunctionTimer(threading.Timer):
    logger = None
    param = None
    entries = None
    diff = None
    dt = None
    fadein = False
    fadeout = False
    switcher = False
    loader = False

    def __init__(self, diff=None, func=None, param=None, fadein=False, fadeout=False, switcher=False, loader=False):

        self.logger = logging.getLogger("AuraEngine")
        self.logger.debug("Executing engine command '%s' in %s seconds..." % (str(func.__name__), str(diff)))
        threading.Timer.__init__(self, diff, func, (param,))

        if not fadein and not fadeout and not switcher and not loader \
            or fadein and fadeout \
            or fadein and switcher \
            or fadeout and switcher:
            
            raise Exception("You have to create me with either fadein=True, fadeout=True or switcher=True")

        self.diff = diff
        self.dt = datetime.now() + timedelta(seconds=diff)

        self.func = func
        self.param = param
        self.entries = param # TODO Refactor since param can hold [entries] or a timeslot, depending on the timer type
        self.fadein = fadein
        self.fadeout = fadeout
        self.switcher = switcher
        self.loader = loader


    def __str__(self):
        """
        String represenation of the timer.
        """
        status = "Timer (Alive: %s)" % self.is_alive()
        status += " starting at " + str(self.dt)

        if self.fadein:
            return status + " fading in entries '" + ResourceUtil.get_entries_string(self.entries)
        elif self.fadeout:
            return status + " fading out timeslot '" + str(self.param)
        elif self.switcher:
            return status + " switching to entries '" + ResourceUtil.get_entries_string(self.entries)
        elif self.loader:
            return status + " preloading entries '" + ResourceUtil.get_entries_string(self.entries)
        else:
            return "CORRUPTED CallFunctionTimer around! How can that be?"
